Intlr <- function(exp, obs, target_grid, int_method, lr_method, predictors = NULL, loocv = FALSE, ncores = 1) {

  require(multiApply) 

  if (!inherits(exp,'s2dv_cube')) {
    stop("Parameter 'exp' must be of the class 's2dv_cube'")
  }
 
  if (!inherits(obs,'s2dv_cube')) {
    stop("Parameter 'obs' must be of the class 's2dv_cube'")
  }

  if (is.array(predictors)) {
    stopifnot('sdate' %in% names(dim(predictors)))
  }

  stopifnot('sdate' %in% names(dim(exp$data)))
  stopifnot('sdate' %in% names(dim(obs$data))) 
  
  lonmin <- attr(exp$lon,"first_lon") # this should not be needed
  lonmax <- attr(exp$lon,"last_lon")
  latmin <- attr(exp$lat,"first_lat")
  latmax <- attr(exp$lat,"last_lat")
  exp_interpolated <- interpolation(exp = exp, target_grid = target_grid, method = int_method,
                                    crop = c(lonmin, lonmax, latmin, latmax))

  # Interpolate obs to the same target grid to ensure the matching with exp coordinates
  obs_interpolated <- interpolation(exp = obs, target_grid = target_grid, method = int_method,
                                    crop = c(lonmin, lonmax, latmin, latmax))

  # TODO: add check .check_coords?
  if (lr_method == 'basic') {
    pred <- exp_interpolated$data
    trgt_dims <- 'sdate'
   } 

  # les observacions són indexos i s'utilitzen de predictors??
  else if (lr_method == 'large-scale') {  
    if (is.null(predictors)) {
      stop("The large-scale predictors must be passed through the parametre 'predictors'")
    }  
    pred <- predictors
    
    var_dim_in_predictors <- names(dim(predictors))[names(dim(predictors)) != 'sdate']
    trgt_dims <- c('sdate', var_dim_in_predictors)
  } 

  else if (lr_method == '4nn') {
    pred <- Apply(list(exp_interpolated$data), target_dims = list(c('lat','lon')), fun = find_4nn)$output1
    trgt_dims <- c('sdate','nn')
  }

  else {
    stop(paste0(lr_method, " method is not implemented yet"))
  }

  res <- Apply(list(pred, obs_interpolated$data), target_dims = list(trgt_dims, 'sdate'), fun = .intlr, 
               loocv = loocv, ncores = ncores)$output1
 
  names(dim(res))[1] <- 'sdate'

  # Reorder dimensions to match those of the input model data
  res <- .reorder_dims(arr_ref = exp$data, arr_to_reorder = res)

  # Create an s2dv_cube object
  res_s2dv <- suppressWarnings(s2dv_cube(data = res, lon = exp_interpolated$lon, lat = exp_interpolated$lat))

  return(res_s2dv)
}

.intlr <- function(x, y, loocv) {

  tmp_df <- data.frame(x = x, y = y)

  # training
  lm1 <- train_lm(df = tmp_df, loocv = loocv)  

  # prediction
  res <- pred_lm(lm1 = lm1, df = tmp_df, loocv = loocv)

  return(res)
 
}

train_lm <- function(df, loocv) {

  # Remove columns containing only NA's
  df <- df[ , apply(df, 2, function(x) !all(is.na(x)))]

  if (loocv) {
   
    lm1 <- lapply(1:nrow(df), function(j) lm(df[-j,], formula = y ~ .))
 
  } else {

    lm1 <- list(lm(data = df, formula = y ~ .))
  }

  return(lm1)
}

pred_lm <- function(df, lm1, loocv) {

  if (loocv) {

    pred_vals <- sapply(1:nrow(df), function(j) predict(lm1[[j]], df[j,]))
  
  } else {
  
    pred_vals_ls <- lapply(lm1, predict, data = df)    
    pred_vals <- unlist(pred_vals_ls)
  } 
 
  return(pred_vals)
}

find_4nn <- function(object) {

  nearests <- InsertDim(object, posdim = 3, lendim = 4, name = 'nn')
  
  for (i in seq(dim(object)['lon'])) {
   lon_max_border <- FALSE
   lon_min_border <- FALSE 
   if (((i + 1) > dim(object)['lon'])) {
      lon_max_border <- TRUE
   } else if (((i - 1) < 1)) {
      lon_min_border <- TRUE
   }
    for (j in seq(dim(object)['lat'])) {
      lat_max_border <- FALSE
      lat_min_border <- FALSE
      if (((j + 1) > dim(object)['lat'])) {
        lat_max_border <- TRUE
      } else if (((j - 1) < 1)) {
        lat_min_border <- TRUE
      }
      if (lon_max_border) {
        nearests[j, i, 2] <- NA
        nearests[j, i, 4] <- object[j, i - 1]
      } else if (lon_min_border) {
        nearests[j, i, 2] <- object[j, i + 1]
        nearests[j, i, 4] <- NA
      } else {
        nearests[j, i, 2] <- object[j, i + 1]
        nearests[j, i, 4] <- object[j, i - 1]
      }
      if (lat_max_border) {
        nearests[j, i, 1] <- NA
        nearests[j, i, 3] <- object[j - 1, i]
      } else if (lat_min_border) {
        nearests[j, i, 1] <- object[j + 1, i]
        nearests[j, i, 3] <- NA
      } else {
        nearests[j, i, 1] <- object[j + 1, i]
        nearests[j, i, 3] <- object[j - 1, i]
      }
    }
  }

  return(nearests)
}


